@@ -31,6 +31,7 @@
 #include <glib.h>
 
 #include <ofono/log.h>
+#include <ofono/modem.h>
 #include "driver.h"
 #include "smsutil.h"
 #include "util.h"
@@ -57,19 +58,28 @@ static gboolean set_cpms(gpointer user_data);
 static const char *storages[] = {
 	"SM",
 	"ME",
+	"MT",
 };
 
 #define SM_STORE 0
 #define ME_STORE 1
+#define MT_STORE 2
 
 struct sms_data {
 	int store;
+	int incoming;
 	int retries;
 	gboolean cnma_enabled;
 	char *cnma_ack_pdu;
 	int cnma_ack_pdu_len;
 };
 
+struct cpms_request {
+	struct ofono_modem *modem;
+	int store;
+	int index;
+};
+
 static struct sms_data *sms_create()
 {
 	return g_try_new0(struct sms_data, 1);
@@ -99,7 +109,7 @@ static void at_csca_set(struct ofono_modem *modem,
 			const struct ofono_phone_number *sca,
 			ofono_generic_cb_t cb, void *data)
 {
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct at_data *at = ofono_modem_get_userdata(modem);
 	struct cb_data *cbd = cb_data_new(modem, cb, data);
 	char buf[64];
 
@@ -174,7 +184,7 @@ err:
 static void at_csca_query(struct ofono_modem *modem, ofono_sca_query_cb_t cb,
 					void *data)
 {
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct at_data *at = ofono_modem_get_userdata(modem);
 	struct cb_data *cbd = cb_data_new(modem, cb, data);
 
 	if (!cbd)
@@ -231,10 +241,10 @@ err:
 }
 
 static void at_cmgs(struct ofono_modem *modem, unsigned char *pdu, int pdu_len,
-			int tpdu_len, gboolean mms, ofono_sms_submit_cb_t cb,
+			int tpdu_len, int mms, ofono_sms_submit_cb_t cb,
 			void *data)
 {
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct at_data *at = ofono_modem_get_userdata(modem);
 	struct cb_data *cbd = cb_data_new(modem, cb, data);
 	char buf[512];
 	int len;
@@ -242,9 +252,11 @@ static void at_cmgs(struct ofono_modem *modem, unsigned char *pdu, int pdu_len,
 	if (!cbd)
 		goto error;
 
-	if (mms)
-		g_at_chat_send(at->parser, "AT+CMMS=1", none_prefix,
+	if (mms) {
+		sprintf(buf, "AT+CMMS=%d", mms);
+		g_at_chat_send(at->parser, buf, none_prefix,
 				NULL, NULL, NULL);
+	}
 
 	len = sprintf(buf, "AT+CMGS=%d\r", tpdu_len);
 	encode_hex_own_buf(pdu, pdu_len, 0, buf+len);
@@ -315,7 +327,7 @@ static void at_cbm_notify(GAtResult *result, gpointer user_data)
 static void at_cds_notify(GAtResult *result, gpointer user_data)
 {
 	struct ofono_modem *modem = user_data;
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct at_data *at = ofono_modem_get_userdata(modem);
 	int pdulen;
 	const char *pdu;
 	char buf[256];
@@ -342,7 +354,7 @@ static void at_cds_notify(GAtResult *result, gpointer user_data)
 static void at_cmt_notify(GAtResult *result, gpointer user_data)
 {
 	struct ofono_modem *modem = user_data;
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct at_data *at = ofono_modem_get_userdata(modem);
 	const char *hexpdu;
 	long pdu_len;
 	int tpdu_len;
@@ -356,7 +368,7 @@ static void at_cmt_notify(GAtResult *result, gpointer user_data)
 		return;
 	}
 
-	ofono_debug("Got new SMS Deliver PDU via CMT: %s, %d", pdu, tpdu_len);
+	ofono_debug("Got new SMS Deliver PDU via CMT: %s, %d", hexpdu, tpdu_len);
 
 	decode_hex_own_buf(hexpdu, -1, &pdu_len, 0, pdu);
 	ofono_sms_deliver_notify(modem, pdu, pdu_len, tpdu_len);
@@ -420,13 +432,36 @@ static void at_cmgd_cb(gboolean ok, GAtResult *result, gpointer user_data)
 		ofono_error("Unable to delete received SMS");
 }
 
+static void at_cmti_cpms_cb(gboolean ok, GAtResult *result, gpointer user_data)
+{
+	struct cpms_request *req = user_data;
+	struct ofono_modem *modem = req->modem;
+	struct at_data *at = ofono_modem_get_userdata(modem);
+	char buf[128];
+
+	if (!ok) {
+		ofono_error("Received CMTI, but CPMS request failed");
+		return;
+	}
+
+	at->sms->store = req->store;
+
+	sprintf(buf, "AT+CMGR=%d", req->index);
+	g_at_chat_send(at->parser, buf, none_prefix, at_cmgr_cb, modem, NULL);
+
+	/* We don't buffer SMS on the SIM/ME, send along a CMGD as well */
+	sprintf(buf, "AT+CMGD=%d", req->index);
+	g_at_chat_send(at->parser, buf, none_prefix, at_cmgd_cb, NULL, NULL);
+}
+
 static void at_cmti_notify(GAtResult *result, gpointer user_data)
 {
 	struct ofono_modem *modem = user_data;
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct at_data *at = ofono_modem_get_userdata(modem);
+	const char *strstore;
+	int store;
 	GAtResultIter iter;
 	int index;
-	char buf[128];
 
 	dump_response("at_cmti_notify", TRUE, result);
 
@@ -435,25 +470,44 @@ static void at_cmti_notify(GAtResult *result, gpointer user_data)
 	if (!g_at_result_iter_next(&iter, "+CMTI:"))
 		goto err;
 
-	/* We skip the memory store here as we assume the modem will honor
-	 * the CPMS we sent out earlier
-	 */
-	if (!g_at_result_iter_skip_next(&iter))
+	if (!g_at_result_iter_next_string(&iter, &strstore))
+		goto err;
+
+	if (!strcmp(strstore, "ME"))
+		store = ME_STORE;
+	else if (!strcmp(strstore, "SM"))
+		store = SM_STORE;
+	else
 		goto err;
 
 	if (!g_at_result_iter_next_number(&iter, &index))
 		goto err;
 
-	ofono_debug("Got a CMTI indication at index: %d", index);
+	ofono_debug("Got a CMTI indication at %s, index: %d", strstore, index);
 
-	sprintf(buf, "AT+CMGR=%d", index);
+	if (store == at->sms->store) {
+		struct cpms_request req;
 
-	g_at_chat_send(at->parser, buf, none_prefix, at_cmgr_cb, modem, NULL);
+		req.modem = modem;
+		req.store = store;
+		req.index = index;
 
-	sprintf(buf, "AT+CMGD=%d", index);
+		at_cmti_cpms_cb(TRUE, NULL, &req);
+	} else {
+		char buf[128];
+		const char *incoming = storages[at->sms->incoming];
+		struct cpms_request *req = g_new(struct cpms_request, 1);
 
-	/* We don't buffer SMS on the SIM/ME, send along a CMGD as well */
-	g_at_chat_send(at->parser, buf, none_prefix, at_cmgd_cb, NULL, NULL);
+		req->modem = modem;
+		req->store = store;
+		req->index = index;
+
+		sprintf(buf, "AT+CPMS=\"%s\",\"%s\",\"%s\"",
+			strstore, strstore, incoming);
+
+		g_at_chat_send(at->parser, buf, cpms_prefix, at_cmti_cpms_cb,
+				req, g_free);
+	}
 
 	return;
 
@@ -463,7 +517,7 @@ err:
 
 static void at_sms_initialized(struct ofono_modem *modem)
 {
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct at_data *at = ofono_modem_get_userdata(modem);
 
 	g_at_chat_register(at->parser, "+CMTI:", at_cmti_notify, FALSE,
 				modem, NULL);
@@ -483,7 +537,7 @@ static void at_sms_initialized(struct ofono_modem *modem)
 
 static void at_sms_not_supported(struct ofono_modem *modem)
 {
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct at_data *at = ofono_modem_get_userdata(modem);
 
 	ofono_error("SMS not supported by this modem.  If this is in error"
 			" please submit patches to support this hardware");
@@ -598,7 +652,7 @@ err:
 static void at_cnmi_query_cb(gboolean ok, GAtResult *result, gpointer user_data)
 {
 	struct ofono_modem *modem = user_data;
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct at_data *at = ofono_modem_get_userdata(modem);
 	GAtResultIter iter;
 	int cnmi_opts[5]; /* See 27.005 Section 3.4.1 */
 	int opt;
@@ -606,6 +660,8 @@ static void at_cnmi_query_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	gboolean supported = FALSE;
 	char buf[128];
 
+	dump_response("at_cnmi_query_cb", ok, result);
+
 	if (!ok)
 		goto out;
 
@@ -617,11 +673,15 @@ static void at_cnmi_query_cb(gboolean ok, GAtResult *result, gpointer user_data)
 		goto out;
 
 	for (opt = 0; opt < 5; opt++) {
+		int min, max;
+
 		if (!g_at_result_iter_open_list(&iter))
 			goto out;
 
-		while (g_at_result_iter_next_number(&iter, &mode))
-			cnmi_opts[opt] |= 1 << mode;
+		while (g_at_result_iter_next_range(&iter, &min, &max)) {
+			for (mode = min; mode <= max; mode++)
+				cnmi_opts[opt] |= 1 << mode;
+		}
 
 		if (!g_at_result_iter_close_list(&iter))
 			goto out;
@@ -643,7 +703,7 @@ out:
 
 static void at_query_cnmi(struct ofono_modem *modem)
 {
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct at_data *at = ofono_modem_get_userdata(modem);
 
 	g_at_chat_send(at->parser, "AT+CNMI=?", cnmi_prefix,
 			at_cnmi_query_cb, modem, NULL);
@@ -652,7 +712,9 @@ static void at_query_cnmi(struct ofono_modem *modem)
 static void at_cpms_set_cb(gboolean ok, GAtResult *result, gpointer user_data)
 {
 	struct ofono_modem *modem = user_data;
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct at_data *at = ofono_modem_get_userdata(modem);
+
+	dump_response("at_cpms_set_cb", ok, result);
 
 	if (ok)
 		return at_query_cnmi(modem);
@@ -670,11 +732,12 @@ static void at_cpms_set_cb(gboolean ok, GAtResult *result, gpointer user_data)
 static gboolean set_cpms(gpointer user_data)
 {
 	struct ofono_modem *modem = user_data;
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct at_data *at = ofono_modem_get_userdata(modem);
 	const char *store = storages[at->sms->store];
+	const char *incoming = storages[at->sms->incoming];
 	char buf[128];
 
-	sprintf(buf, "AT+CPMS=\"%s\",\"%s\",\"%s\"", store, store, store);
+	sprintf(buf, "AT+CPMS=\"%s\",\"%s\",\"%s\"", store, store, incoming);
 
 	g_at_chat_send(at->parser, buf, cpms_prefix,
 			at_cpms_set_cb, modem, NULL);
@@ -684,7 +747,9 @@ static gboolean set_cpms(gpointer user_data)
 static void at_cmgf_set_cb(gboolean ok, GAtResult *result, gpointer user_data)
 {
 	struct ofono_modem *modem = user_data;
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct at_data *at = ofono_modem_get_userdata(modem);
+
+	dump_response("at_cmgf_set_cb", ok, result);
 
 	if (ok) {
 		at->sms->retries = 0;
@@ -705,7 +770,7 @@ static void at_cmgf_set_cb(gboolean ok, GAtResult *result, gpointer user_data)
 static gboolean set_cmgf(gpointer user_data)
 {
 	struct ofono_modem *modem = user_data;
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct at_data *at = ofono_modem_get_userdata(modem);
 
 	g_at_chat_send(at->parser, "AT+CMGF=0", cmgf_prefix,
 			at_cmgf_set_cb, modem, NULL);
@@ -716,18 +781,22 @@ static void at_cpms_query_cb(gboolean ok, GAtResult *result,
 				gpointer user_data)
 {
 	struct ofono_modem *modem = user_data;
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct at_data *at = ofono_modem_get_userdata(modem);
 	gboolean supported = FALSE;
 
+	dump_response("cpms_query_cb", ok, result);
+
 	if (ok) {
 		int mem = 0;
 		GAtResultIter iter;
 		const char *store;
 		gboolean me_supported[3];
 		gboolean sm_supported[3];
+		gboolean mt_supported[3];
 
 		memset(me_supported, 0, sizeof(me_supported));
 		memset(sm_supported, 0, sizeof(sm_supported));
+		memset(mt_supported, 0, sizeof(mt_supported));
 
 		g_at_result_iter_init(&iter, result);
 
@@ -743,21 +812,39 @@ static void at_cpms_query_cb(gboolean ok, GAtResult *result,
 					me_supported[mem] = TRUE;
 				else if (!strcmp(store, "SM"))
 					sm_supported[mem] = TRUE;
+				else if (!strcmp(store, "MT"))
+					mt_supported[mem] = TRUE;
 			}
 
 			if (!g_at_result_iter_close_list(&iter))
 				goto out;
 		}
 
-		if (me_supported[0] && me_supported[1] && me_supported[2]) {
+		if (!sm_supported[2] && !me_supported[2] && !mt_supported[2])
+			goto out;
+
+		if (sm_supported[0] && sm_supported[1]) {
+			supported = TRUE;
+			at->sms->store = SM_STORE;
+		}
+
+		if (me_supported[0] && me_supported[1]) {
 			supported = TRUE;
 			at->sms->store = ME_STORE;
 		}
 
-		if (sm_supported[0] && sm_supported[1] && sm_supported[2]) {
-			supported = TRUE;
-			at->sms->store = SM_STORE;
-		}
+		/* This seems to be a special case, where the modem will
+		 * pick & route the SMS to any of the storages supported by
+		 * mem1
+		 */
+		if (mt_supported[2] && (sm_supported[0] || me_supported[0]))
+			at->sms->incoming = MT_STORE;
+
+		if (sm_supported[2])
+			at->sms->incoming = SM_STORE;
+
+		if (me_supported[2])
+			at->sms->incoming = ME_STORE;
 	}
 out:
 	if (!supported)
@@ -770,9 +857,11 @@ static void at_cmgf_query_cb(gboolean ok, GAtResult *result,
 				gpointer user_data)
 {
 	struct ofono_modem *modem = user_data;
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct at_data *at = ofono_modem_get_userdata(modem);
 	gboolean supported = FALSE;
 
+	dump_response("cmgf_query_cb", ok, result);
+
 	if (ok) {
 		GAtResultIter iter;
 		int mode;
@@ -785,8 +874,9 @@ static void at_cmgf_query_cb(gboolean ok, GAtResult *result,
 		if (!g_at_result_iter_open_list(&iter))
 			goto out;
 
+		/* Look for mode 0 (PDU mode) */
 		while (g_at_result_iter_next_number(&iter, &mode))
-			if (mode == 1)
+			if (mode == 0)
 				supported = TRUE;
 	}
 
@@ -802,9 +892,11 @@ static void at_csms_status_cb(gboolean ok, GAtResult *result,
 				gpointer user_data)
 {
 	struct ofono_modem *modem = user_data;
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct at_data *at = ofono_modem_get_userdata(modem);
 	gboolean supported = FALSE;
 
+	dump_response("csms_status_cb", ok, result);
+
 	if (ok) {
 		GAtResultIter iter;
 		int service;
@@ -845,7 +937,7 @@ static void at_csms_set_cb(gboolean ok, GAtResult *result,
 				gpointer user_data)
 {
 	struct ofono_modem *modem = user_data;
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct at_data *at = ofono_modem_get_userdata(modem);
 
 	g_at_chat_send(at->parser, "AT+CSMS?", csms_prefix,
 			at_csms_status_cb, modem, NULL);
@@ -855,12 +947,14 @@ static void at_csms_query_cb(gboolean ok, GAtResult *result,
 				gpointer user_data)
 {
 	struct ofono_modem *modem = user_data;
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct at_data *at = ofono_modem_get_userdata(modem);
 	gboolean cnma_supported = FALSE;
 	GAtResultIter iter;
 	int status;
 	char buf[128];
 
+	dump_response("csms_query_cb", ok, result);
+
 	if (!ok)
 		return at_sms_not_supported(modem);
 
@@ -891,7 +985,7 @@ out:
 
 void at_sms_init(struct ofono_modem *modem)
 {
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct at_data *at = ofono_modem_get_userdata(modem);
 
 	g_at_chat_send(at->parser, "AT+CSMS=?", csms_prefix,
 			at_csms_query_cb, modem, NULL);
@@ -899,7 +993,7 @@ void at_sms_init(struct ofono_modem *modem)
 
 void at_sms_exit(struct ofono_modem *modem)
 {
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct at_data *at = ofono_modem_get_userdata(modem);
 
 	if (!at->sms)
 		return;
