@@ -29,34 +29,17 @@
 #include <string.h>
 #include <assert.h>
 #include <termios.h>
+#include <ctype.h>
 
 #include <glib.h>
 
 #include "ringbuffer.h"
-#include "gatresult.h"
 #include "gatchat.h"
 
 /* #define WRITE_SCHEDULER_DEBUG 1 */
 
 static void g_at_chat_wakeup_writer(GAtChat *chat);
-
-enum chat_state {
-	PARSER_STATE_IDLE = 0,
-	PARSER_STATE_INITIAL_CR,
-	PARSER_STATE_INITIAL_LF,
-	PARSER_STATE_RESPONSE,
-	PARSER_STATE_TERMINATOR_CR,
-	PARSER_STATE_RESPONSE_COMPLETE,
-	PARSER_STATE_GUESS_MULTILINE_RESPONSE,
-	PARSER_STATE_MULTILINE_RESPONSE,
-	PARSER_STATE_MULTILINE_TERMINATOR_CR,
-	PARSER_STATE_MULTILINE_COMPLETE,
-	PARSER_STATE_PDU,
-	PARSER_STATE_PDU_CR,
-	PARSER_STATE_PDU_COMPLETE,
-	PARSER_STATE_PROMPT,
-	PARSER_STATE_PROMPT_COMPLETE
-};
+static void debug_chat(GAtChat *chat, gboolean in, const char *str, gsize len);
 
 struct at_command {
 	char *cmd;
@@ -95,14 +78,16 @@ struct _GAtChat {
 	struct ring_buffer *buf;		/* Current read buffer */
 	guint read_so_far;			/* Number of bytes processed */
 	gboolean disconnecting;			/* Whether we're disconnecting */
-	enum chat_state state;		/* Current chat state */
-	int flags;
+	GAtDebugFunc debugf;			/* debugging output function */
+	gpointer debug_data;			/* Data to pass to debug func */
 	char *pdu_notify;			/* Unsolicited Resp w/ PDU */
 	GSList *response_lines;			/* char * lines of the response */
 	char *wakeup;				/* command sent to wakeup modem */
+	gint timeout_source;
 	gdouble inactivity_time;		/* Period of inactivity */
 	guint wakeup_timeout;			/* How long to wait for resp */
 	GTimer *wakeup_timer;			/* Keep track of elapsed time */
+	GAtSyntax *syntax;
 };
 
 static gint at_notify_node_compare_by_id(gconstpointer a, gconstpointer b)
@@ -254,6 +239,9 @@ static void g_at_chat_cleanup(GAtChat *chat)
 		g_timer_destroy(chat->wakeup_timer);
 		chat->wakeup_timer = 0;
 	}
+
+	g_at_syntax_unref(chat->syntax);
+	chat->syntax = NULL;
 }
 
 static void read_watcher_destroy_notify(GAtChat *chat)
@@ -306,7 +294,10 @@ static gboolean g_at_chat_match_notify(GAtChat *chat, char *line)
 
 		if (notify->pdu) {
 			chat->pdu_notify = line;
-			chat->state = PARSER_STATE_PDU;
+
+			if (chat->syntax->set_hint)
+				chat->syntax->set_hint(chat->syntax,
+							G_AT_SYNTAX_EXPECT_PDU);
 			return TRUE;
 		}
 
@@ -321,7 +312,6 @@ static gboolean g_at_chat_match_notify(GAtChat *chat, char *line)
 	if (ret) {
 		g_slist_free(result.lines);
 		g_free(line);
-		chat->state = PARSER_STATE_IDLE;
 	}
 
 	return ret;
@@ -387,8 +377,6 @@ static gboolean g_at_chat_handle_command_response(GAtChat *p,
 	int i;
 	int size = sizeof(terminator_table) / sizeof(struct terminator_info);
 
-	p->state = PARSER_STATE_IDLE;
-
 	for (i = 0; i < size; i++) {
 		struct terminator_info *info = &terminator_table[i];
 
@@ -415,8 +403,8 @@ static gboolean g_at_chat_handle_command_response(GAtChat *p,
 	}
 
 out:
-	if (!(p->flags & G_AT_CHAT_FLAG_NO_LEADING_CRLF))
-		p->state = PARSER_STATE_GUESS_MULTILINE_RESPONSE;
+	if (p->syntax->set_hint)
+		p->syntax->set_hint(p->syntax, G_AT_SYNTAX_EXPECT_MULTILINE);
 
 	if (cmd->listing) {
 		GAtResult result;
@@ -434,31 +422,13 @@ out:
 	return TRUE;
 }
 
-static void have_line(GAtChat *p, gboolean strip_preceding)
+static void have_line(GAtChat *p, char *str)
 {
 	/* We're not going to copy terminal <CR><LF> */
-	unsigned int len = p->read_so_far - 2;
-	char *str;
 	struct at_command *cmd;
 
-	/* If we have preceding <CR><LF> modify the len */
-	if (strip_preceding)
-		len -= 2;
-
-	/* Make sure we have terminal null */
-	str = g_try_new(char, len + 1);
-
-	if (!str) {
-		ring_buffer_drain(p->buf, p->read_so_far);
+	if (!str)
 		return;
-	}
-
-	if (strip_preceding)
-		ring_buffer_drain(p->buf, 2);
-	ring_buffer_read(p->buf, str, len);
-	ring_buffer_drain(p->buf, 2);
-
-	str[len] = '\0';
 
 	/* Check for echo, this should not happen, but lets be paranoid */
 	if (!strncmp(str, "AT", 2) == TRUE)
@@ -466,7 +436,7 @@ static void have_line(GAtChat *p, gboolean strip_preceding)
 
 	cmd = g_queue_peek_head(p->command_queue);
 
-	if (cmd) {
+	if (cmd && p->cmd_bytes_written > 0) {
 		char c = cmd->cmd[p->cmd_bytes_written - 1];
 
 		/* We check that we have submitted a terminator, in which case
@@ -488,30 +458,18 @@ static void have_line(GAtChat *p, gboolean strip_preceding)
 done:
 	/* No matches & no commands active, ignore line */
 	g_free(str);
-	p->state = PARSER_STATE_IDLE;
 }
 
-static void have_pdu(GAtChat *p)
+static void have_pdu(GAtChat *p, char *pdu)
 {
-	unsigned int len = p->read_so_far - 2;
-	char *pdu;
 	GHashTableIter iter;
 	struct at_notify *notify;
 	char *prefix;
 	gpointer key, value;
 	GAtResult result;
 
-	pdu = g_try_new(char, len + 1);
-
-	if (!pdu) {
-		ring_buffer_drain(p->buf, p->read_so_far);
+	if (!pdu)
 		goto out;
-	}
-
-	ring_buffer_read(p->buf, pdu, len);
-	ring_buffer_drain(p->buf, 2);
-
-	pdu[len] = '\0';
 
 	result.lines = g_slist_prepend(NULL, p->pdu_notify);
 	result.final_or_pdu = pdu;
@@ -540,97 +498,47 @@ out:
 
 	if (pdu)
 		g_free(pdu);
-
-	p->state = PARSER_STATE_IDLE;
 }
 
-static inline void parse_char(GAtChat *chat, char byte)
+static char *extract_line(GAtChat *p)
 {
-	switch (chat->state) {
-	case PARSER_STATE_IDLE:
-		if (byte == '\r')
-			chat->state = PARSER_STATE_INITIAL_CR;
-		else if (chat->flags & G_AT_CHAT_FLAG_NO_LEADING_CRLF) {
-			if (byte == '>')
-				chat->state = PARSER_STATE_PROMPT;
+	unsigned int wrap = ring_buffer_len_no_wrap(p->buf);
+	unsigned int pos = 0;
+	unsigned char *buf = ring_buffer_read_ptr(p->buf, pos);
+	int strip_front = 0;
+	int line_length = 0;
+	char *line;
+
+	while (pos < p->read_so_far) {
+		if (*buf == '\r' || *buf == '\n')
+			if (!line_length)
+				strip_front += 1;
 			else
-				chat->state = PARSER_STATE_RESPONSE;
-		}
-		break;
-
-	case PARSER_STATE_INITIAL_CR:
-		if (byte == '\n')
-			chat->state = PARSER_STATE_INITIAL_LF;
-		else if (byte != '\r' && /* Echo & no <CR><LF>?! */
-			(chat->flags & G_AT_CHAT_FLAG_NO_LEADING_CRLF))
-			chat->state = PARSER_STATE_RESPONSE;
-		else if (byte != '\r')
-			chat->state = PARSER_STATE_IDLE;
-		break;
-
-	case PARSER_STATE_INITIAL_LF:
-		if (byte == '\r')
-			chat->state = PARSER_STATE_TERMINATOR_CR;
-		else if (byte == '>')
-			chat->state = PARSER_STATE_PROMPT;
+				break;
 		else
-			chat->state = PARSER_STATE_RESPONSE;
-		break;
+			line_length += 1;
 
-	case PARSER_STATE_RESPONSE:
-		if (byte == '\r')
-			chat->state = PARSER_STATE_TERMINATOR_CR;
-		break;
+		buf += 1;
+		pos += 1;
 
-	case PARSER_STATE_TERMINATOR_CR:
-		if (byte == '\n')
-			chat->state = PARSER_STATE_RESPONSE_COMPLETE;
-		else
-			chat->state = PARSER_STATE_IDLE;
-		break;
-
-	case PARSER_STATE_GUESS_MULTILINE_RESPONSE:
-		if (byte == '\r')
-			chat->state = PARSER_STATE_INITIAL_CR;
-		else
-			chat->state = PARSER_STATE_MULTILINE_RESPONSE;
-		break;
-
-	case PARSER_STATE_MULTILINE_RESPONSE:
-		if (byte == '\r')
-			chat->state = PARSER_STATE_MULTILINE_TERMINATOR_CR;
-		break;
-
-	case PARSER_STATE_MULTILINE_TERMINATOR_CR:
-		if (byte == '\n')
-			chat->state = PARSER_STATE_MULTILINE_COMPLETE;
-		break;
-
-	case PARSER_STATE_PDU:
-		if (byte == '\r')
-			chat->state = PARSER_STATE_PDU_CR;
-		break;
-
-	case PARSER_STATE_PDU_CR:
-		if (byte == '\n')
-			chat->state = PARSER_STATE_PDU_COMPLETE;
-		break;
-
-	case PARSER_STATE_PROMPT:
-		if (byte == ' ')
-			chat->state = PARSER_STATE_PROMPT_COMPLETE;
-		else
-			chat->state = PARSER_STATE_RESPONSE;
-		break;
-
-	case PARSER_STATE_RESPONSE_COMPLETE:
-	case PARSER_STATE_PDU_COMPLETE:
-	case PARSER_STATE_MULTILINE_COMPLETE:
-	default:
-		/* This really shouldn't happen */
-		assert(FALSE);
-		return;
+		if (pos == wrap)
+			buf = ring_buffer_read_ptr(p->buf, pos);
 	}
+
+	line = g_try_new(char, line_length + 1);
+
+	if (!line) {
+		ring_buffer_drain(p->buf, p->read_so_far);
+		return NULL;
+	}
+
+	ring_buffer_drain(p->buf, strip_front);
+	ring_buffer_read(p->buf, line, line_length);
+	ring_buffer_drain(p->buf, p->read_so_far - strip_front - line_length);
+
+	line[line_length] = '\0';
+
+	return line;
 }
 
 static void new_bytes(GAtChat *p)
@@ -639,70 +547,126 @@ static void new_bytes(GAtChat *p)
 	unsigned int wrap = ring_buffer_len_no_wrap(p->buf);
 	unsigned char *buf = ring_buffer_read_ptr(p->buf, p->read_so_far);
 
-	while (p->read_so_far < len) {
-		parse_char(p, *buf);
+	GAtSyntaxResult result;
 
-		buf += 1;
-		p->read_so_far += 1;
+	while (p->read_so_far < len) {
+		gsize rbytes = MIN(len - p->read_so_far, wrap - p->read_so_far);
+		result = p->syntax->feed(p->syntax, (char *)buf, &rbytes);
+
+		buf += rbytes;
+		p->read_so_far += rbytes;
 
 		if (p->read_so_far == wrap) {
 			buf = ring_buffer_read_ptr(p->buf, p->read_so_far);
 			wrap = len;
 		}
 
-		if (p->state == PARSER_STATE_RESPONSE_COMPLETE) {
-			gboolean strip_preceding;
+		if (result == G_AT_SYNTAX_RESULT_UNSURE)
+			continue;
 
-			if (p->flags & G_AT_CHAT_FLAG_NO_LEADING_CRLF)
-				strip_preceding = FALSE;
-			else
-				strip_preceding = TRUE;
+		switch (result) {
+		case G_AT_SYNTAX_RESULT_LINE:
+		case G_AT_SYNTAX_RESULT_MULTILINE:
+			have_line(p, extract_line(p));
+			break;
 
-			len -= p->read_so_far;
-			wrap -= p->read_so_far;
-
-			have_line(p, strip_preceding);
-
-			p->read_so_far = 0;
-		} else if (p->state == PARSER_STATE_MULTILINE_COMPLETE) {
-			len -= p->read_so_far;
-			wrap -= p->read_so_far;
-
-			have_line(p, FALSE);
-
-			p->read_so_far = 0;
-		} else if (p->state == PARSER_STATE_PDU_COMPLETE) {
-			len -= p->read_so_far;
-			wrap -= p->read_so_far;
-
-			have_pdu(p);
-
-			p->read_so_far = 0;
-		} else if (p->state == PARSER_STATE_INITIAL_CR) {
-			len -= p->read_so_far - 1;
-			wrap -= p->read_so_far - 1;
-
-			ring_buffer_drain(p->buf, p->read_so_far - 1);
-
-			p->read_so_far = 1;
-		} else if (p->state == PARSER_STATE_PROMPT_COMPLETE) {
-			len -= p->read_so_far;
-			wrap -= p->read_so_far;
+		case G_AT_SYNTAX_RESULT_PDU:
+			have_pdu(p, extract_line(p));
+			break;
 
+		case G_AT_SYNTAX_RESULT_PROMPT:
 			g_at_chat_wakeup_writer(p);
-
 			ring_buffer_drain(p->buf, p->read_so_far);
+			break;
 
-			p->read_so_far = 0;
+		default:
+			ring_buffer_drain(p->buf, p->read_so_far);
+			break;
+		}
 
-			p->state = PARSER_STATE_IDLE;
+		len -= p->read_so_far;
+		wrap -= p->read_so_far;
+		p->read_so_far = 0;
+	}
+
+	/* We're overflowing the buffer, shutdown the socket */
+	if (ring_buffer_avail(p->buf) == 0)
+		g_at_chat_shutdown(p);
+}
+
+static void debug_chat(GAtChat *chat, gboolean in, const char *str, gsize len)
+{
+	char type = in ? '<' : '>';
+	gsize escaped = 2; /* Enough for '<', ' ' */
+	char *escaped_str;
+	const char *esc = "<ESC>";
+	gsize esc_size = strlen(esc);
+	const char *ctrlz = "<CtrlZ>";
+	gsize ctrlz_size = strlen(ctrlz);
+	gsize i;
+
+	if (!chat->debugf || !len)
+		return;
+
+	for (i = 0; i < len; i++) {
+		char c = str[i];
+
+		if (isprint(c))
+			escaped += 1;
+		else if (c == '\r' || c == '\t' || c == '\n')
+			escaped += 2;
+		else if (c == 26)
+			escaped += ctrlz_size;
+		else if (c == 25)
+			escaped += esc_size;
+		else
+			escaped += 4;
+	}
+
+	escaped_str = g_malloc(escaped + 1);
+	escaped_str[0] = type;
+	escaped_str[1] = ' ';
+	escaped_str[2] = '\0';
+	escaped_str[escaped] = '\0';
+
+	for (escaped = 2, i = 0; i < len; i++) {
+		char c = str[i];
+
+		switch (c) {
+		case '\r':
+			escaped_str[escaped++] = '\\';
+			escaped_str[escaped++] = 'r';
+			break;
+		case '\t':
+			escaped_str[escaped++] = '\\';
+			escaped_str[escaped++] = 't';
+			break;
+		case '\n':
+			escaped_str[escaped++] = '\\';
+			escaped_str[escaped++] = 'n';
+			break;
+		case 26:
+			strncpy(&escaped_str[escaped], ctrlz, ctrlz_size);
+			escaped += ctrlz_size;
+			break;
+		case 25:
+			strncpy(&escaped_str[escaped], esc, esc_size);
+			escaped += esc_size;
+			break;
+		default:
+			if (isprint(c))
+				escaped_str[escaped++] = c;
+			else {
+				escaped_str[escaped++] = '\\';
+				escaped_str[escaped++] = '0' + ((c >> 6) & 07);
+				escaped_str[escaped++] = '0' + ((c >> 3) & 07);
+				escaped_str[escaped++] = '0' + (c & 07);
+			}
 		}
 	}
 
-	if (p->state == PARSER_STATE_IDLE && p->read_so_far > 0) {
-		ring_buffer_drain(p->buf, p->read_so_far);
-		p->read_so_far = 0;
-	}
+	chat->debugf(escaped_str, chat->debug_data);
+	g_free(escaped_str);
 }
 
 static gboolean received_data(GIOChannel *channel, GIOCondition cond,
@@ -724,21 +688,13 @@ static gboolean received_data(GIOChannel *channel, GIOCondition cond,
 
 		toread = ring_buffer_avail_no_wrap(chat->buf);
 
-		/* We're going to start overflowing the buffer
-		 * this cannot happen under normal circumstances, so probably
-		 * the channel is getting garbage, drop off
-		 */
-		if (toread == 0) {
-			if (chat->state == PARSER_STATE_RESPONSE)
-				return FALSE;
-
-			err = G_IO_ERROR_AGAIN;
+		if (toread == 0)
 			break;
-		}
 
 		buf = ring_buffer_write_ptr(chat->buf);
 
 		err = g_io_channel_read(channel, (char *) buf, toread, &rbytes);
+		debug_chat(chat, TRUE, (char *)buf, rbytes);
 
 		total_read += rbytes;
 
@@ -753,9 +709,6 @@ static gboolean received_data(GIOChannel *channel, GIOCondition cond,
 	if (cond & (G_IO_HUP | G_IO_ERR))
 		return FALSE;
 
-	if (err == G_IO_ERROR_NONE && rbytes == 0)
-		return FALSE;
-
 	if (err != G_IO_ERROR_NONE && err != G_IO_ERROR_AGAIN)
 		return FALSE;
 
@@ -765,6 +718,16 @@ static gboolean received_data(GIOChannel *channel, GIOCondition cond,
 static gboolean wakeup_no_response(gpointer user)
 {
 	GAtChat *chat = user;
+	struct at_command *cmd = g_queue_peek_head(chat->command_queue);
+
+	chat->timeout_source = 0;
+
+	/* Sometimes during startup the modem is still in the ready state
+	 * and might acknowledge our 'wakeup' command.  In that case don't
+	 * timeout the wrong command
+	 */
+	if (cmd == NULL || cmd->id != 0)
+		return FALSE;
 
 	g_at_chat_finish_command(chat, FALSE, NULL);
 
@@ -828,8 +791,8 @@ static gboolean can_write_data(GIOChannel *channel, GIOCondition cond,
 
 		len = strlen(chat->wakeup);
 
-		g_timeout_add(chat->wakeup_timeout, wakeup_no_response,
-				chat);
+		chat->timeout_source = g_timeout_add(chat->wakeup_timeout,
+						wakeup_no_response, chat);
 	}
 
 	towrite = len - chat->cmd_bytes_written;
@@ -860,6 +823,8 @@ static gboolean can_write_data(GIOChannel *channel, GIOCondition cond,
 		return FALSE;
 	}
 
+	debug_chat(chat, FALSE, cmd->cmd + chat->cmd_bytes_written,
+			bytes_written);
 	chat->cmd_bytes_written += bytes_written;
 
 	if (bytes_written < towrite)
@@ -884,7 +849,7 @@ static void g_at_chat_wakeup_writer(GAtChat *chat)
 				(GDestroyNotify)write_watcher_destroy_notify);
 }
 
-GAtChat *g_at_chat_new(GIOChannel *channel, int flags)
+GAtChat *g_at_chat_new(GIOChannel *channel, GAtSyntax *syntax)
 {
 	GAtChat *chat;
 	GIOFlags io_flags;
@@ -892,14 +857,18 @@ GAtChat *g_at_chat_new(GIOChannel *channel, int flags)
 	if (!channel)
 		return NULL;
 
+	if (!syntax)
+		return NULL;
+
 	chat = g_try_new0(GAtChat, 1);
 
 	if (!chat)
 		return chat;
 
+	chat->ref_count = 1;
 	chat->next_cmd_id = 1;
 	chat->next_notify_id = 1;
-	chat->flags = flags;
+	chat->debugf = NULL;
 
 	chat->buf = ring_buffer_new(4096);
 
@@ -934,6 +903,8 @@ GAtChat *g_at_chat_new(GIOChannel *channel, int flags)
 				received_data, chat,
 				(GDestroyNotify)read_watcher_destroy_notify);
 
+	chat->syntax = g_at_syntax_ref(syntax);
+
 	return chat;
 
 error:
@@ -970,7 +941,7 @@ static int open_device(const char *device)
 	return fd;
 }
 
-GAtChat *g_at_chat_new_from_tty(const char *device, int flags)
+GAtChat *g_at_chat_new_from_tty(const char *device, GAtSyntax *syntax)
 {
 	GIOChannel *channel;
 	int fd;
@@ -985,7 +956,7 @@ GAtChat *g_at_chat_new_from_tty(const char *device, int flags)
 		return NULL;
 	}
 
-	return g_at_chat_new(channel, flags);
+	return g_at_chat_new(channel, syntax);
 }
 
 GAtChat *g_at_chat_ref(GAtChat *chat)
@@ -1020,6 +991,11 @@ gboolean g_at_chat_shutdown(GAtChat *chat)
 	if (chat->channel == NULL)
 		return FALSE;
 
+	if (chat->timeout_source) {
+		g_source_remove(chat->timeout_source);
+		chat->timeout_source = 0;
+	}
+
 	chat->disconnecting = TRUE;
 
 	if (chat->read_watch)
@@ -1043,6 +1019,17 @@ gboolean g_at_chat_set_disconnect_function(GAtChat *chat,
 	return TRUE;
 }
 
+gboolean g_at_chat_set_debug(GAtChat *chat, GAtDebugFunc func, gpointer user)
+{
+	if (chat == NULL)
+		return FALSE;
+
+	chat->debugf = func;
+	chat->debug_data = user;
+
+	return TRUE;
+}
+
 static guint send_common(GAtChat *chat, const char *cmd,
 			const char **prefix_list,
 			GAtNotifyFunc listing, GAtResultFunc func,
